#include "RT_HW_MCP23017.h"
//---------------------------------------------------------------------------------------
//     МАКРОСЫ ДЛЯ РАБОТЫ С УСТРОЙСТВОМ
//---------------------------------------------------------------------------------------
#define DEVICE_BEGIN_PERIOD_EVENT			50000 	//--Период событий (тики) в mcs, генерируемых для использования отсчета времени в блоках выдачи
#define DEVICE_PERIOD_TIC_DEFAULT			10  	//--Период в тиках по умолчанию
#define DEVICE_I2C_DELAY_RESTART_SHORT		50  	//--Пауза(мс)  для контрольной проверки наличия устройства
#define DEVICE_I2C_DELAY_RESTART_MIDDLE		100   	//--Пауза(мс)  межу попытками рестарта устройства при поиске
#define DEVICE_I2C_DELAY_RESTART_LONG		1000	//--Пауза(мс)  между поиском устройства, если устройства потеряно
#define DEVICE_I2C_QNT_RESTART				4   	//--Кол-во попыток рестарта устройства, если ,был сбой связи
#define DEVICE_MAX_TIME_READ      				2		//--Максимальное время на считывание регистров
//-------------------------------------------------------------------------------------------------
//          Параметры - адреса для IOCON.BANK=0
//-------------------------------------------------------------------------------------------------
#define MCP_BASE	0x20	//начало адресного пространства MCP23017 (на шине i2c)
#define MCP_IODIRA	0x00	//регистр направления пинов A: =0x00(default) - все на выход, 0xff - все на вход
#define MCP_IODIRB	0x01	//регистр направления пинов B: =0x00(default) - все на выход, 0xff - все на вход
#define MCP_IPOLA	0x02	//регистр инверсии    пинов A: =0x00(default) - все прямые,   0xff - все инверсные
#define MCP_IPOLB	0x03	//регистр инверсии    выводов регистра B: =0x00(default) - все прямые,   0xff - все инверсные
#define MCP_GPPUA	0x0C	//регистр подтяжки входов регистра A резисторами 100кОм: =0x00(default) - не подтянуты,   0xff - все подтянуты
#define MCP_GPPUB	0x0D	//регистр подтяжки входов регистра A резисторами 100кОм: =0x00(default) - не подтянуты,   0xff - все подтянуты
#define MCP_GPIOA	0x12	//регистр выводов  на пины A
#define MCP_GPIOB	0x13	//регистр выводов  на пины B
#define MCP_IOCON   0x0A	//регистр конфигурации
//-------------------------------------------------------------------------------------------------
// 			              КОНСТРУКТОР
// 1)Устанавливает параметры по умолчанию. 2)Иницирует параметры устройства zb. 3)Заполняет массив указателей указателями на устройство zb.
//-------------------------------------------------------------------------------------------------
//RT_HW_MCP23017::RT_HW_MCP23017(){initDevice=0; zb.num=0; valA,valB=0; zb.addr=0x20; bitSet(initDevice,0); for(int i=0; i<RT_HW_MCP23017_DEVICE_MAX;i++) {prt[i]=&zb;}};  
//-------------------------------------------------------------------------------------------------
// 			        УСТАНОВКА ID УСТРОЙСТВА  
//-------------------------------------------------------------------------------------------------
void RT_HW_MCP23017:: setID_device(RT_HW_MCP23017_DEVICE &z, int16_t addr, char modeA, char modeB, int16_t period){
  if(addr<0){z.dir=1; return;}
  if(z.dir) {return;} if((addr>=0) && (addr<=7)){z.addr=MCP_BASE+addr;} else {z.addr=addr;}	//--Выход, если инциализация состоялась. Установка адреса: если адрес от 0-7, то это адрес в адресном пространстве MCP_BASE 
  //--Установка флагов настройки порта A
  switch(modeA){default:
	case 'U':z.dirA=1;z.gppuA=1;z.ipolA=1; break; 
	case 'u':z.dirA=1;z.gppuA=1;z.ipolA=0; break; 
	case 'I':z.dirA=1;z.gppuA=0;z.ipolA=1; break; 
	case 'i':z.dirA=1;z.gppuA=0;z.ipolA=0; break;
	case 'O':z.dirA=0;z.gppuA=0;z.ipolA=0; break;
	case 'o':z.dirA=0;z.gppuA=0;z.ipolA=1; break;
	}
  //--Установка флагов настройки порта B	
  switch(modeB){default:
	case 'U':z.dirB=1;z.gppuB=1;z.ipolB=1; break; 
	case 'u':z.dirB=1;z.gppuB=1;z.ipolB=0; break; 
	case 'I':z.dirB=1;z.gppuB=0;z.ipolB=1; break; 
	case 'i':z.dirB=1;z.gppuB=0;z.ipolB=0; break;
	case 'O':z.dirB=0;z.gppuB=0;z.ipolB=0; break;
	case 'o':z.dirB=0;z.gppuB=0;z.ipolB=1; break;	
	} 
  //--Установка флагов разрешения записи	
  if( z.dirA ||  z.dirB) {z.readYes=1;}  else {z.readYes=0;}
  if(!z.dirA || !z.dirB) {z.writeYes=1;} else {z.writeYes=0;}
  //--Установка периода опроса в тиках
  if(period<0) {z.period=DEVICE_PERIOD_TIC_DEFAULT;} else {z.period=period;}
  //--Очистка рабочих регистров
  z.stepBegin=z.count=z.event=0; 
  //--Установка параметров системы
  z.device=0; z.run=3; z.dir=2;
} //--End	//--Очистка рабочих регистров.Установка начального состояния регистра управления устройтвом		
//-------------------------------------------------------------------------------------------------
//            	ИНИЦИАЛИЗАЦИЯ И КОНТРОЛЬ УСТРОЙСТВА
//-------------------------------------------------------------------------------------------------
bool RT_HW_MCP23017:: begin(RT_HW_MCP23017_DEVICE &z){
switch(z.stepBegin){default: z.stepBegin=0;						
  //----ОЖИДАНИЕ УСТАНОВКМ ПАРАМЕТРОВ УСТРОЙСТВА (выполняется функцией setID_device() )О
  case 0: switch(z.dir){default: z.stepBegin=0; case 0: return 0; case 1: return 1; case 2: z.stepBegin++; break;}
  case 1: if(z.device==3) {if(event_mcs(z.timeStartEvent,DEVICE_BEGIN_PERIOD_EVENT)) {z.event=1;} else {z.event=0;} return 1;} //--Генераторо событий
          z.event=0; z.stepFind=0; z.stepBegin++;											//--Переход после обнаружения устройства 			
  //----ПОИСК УСТРОЙСТВА НА ШИНЕ i2c 
  case 2: find(z.stepFind, z.addr, z.device, z.count, z.timeStart); if(z.device!=2) break;	//--Поиск устройства
		  z.timeStart=millis(); z.stepBegin++; break;										//--Переход после обнаружения устройства
  //----ИНИЦИАЛИЗАЦИЯ НАПРАВЛЕНИЯ И ПОЛЯРНОСТИ(для ввода) РЕГИСТРОВ 
  case 3: if(!event_ms(z.timeStart,DEVICE_I2C_DELAY_RESTART_SHORT)) break;
		  z.data[0x00]=MCP_IODIRA; 							//--Адрес регистра, с которого будут записаны четыре параметра
		  if(z.dirA)	z.data[1]=0xff;	else z.data[1]=0;	//--Установка направление регистра A (ввод/вывод)
		  if(z.dirB) 	z.data[2]=0xff;	else z.data[2]=0;	//--Установка направление регистра В (ввод/вывод)
		  if(z.ipolA) 	z.data[3]=0xff;	else z.data[3]=0;	//--Установка полярности регистров A (при работк на ввод)  
		  if(z.ipolB) 	z.data[4]=0xff;	else z.data[4]=0;	//--Установка полярности регистров A (при работк на ввод)  	
		  Wire.beginTransmission(z.addr); Wire.write(z.data,5);	//--Передача в i2c-slave 5 байт.
		  if(Wire.endTransmission()) {z.device=z.stepBegin=0; return 1;}  z.stepBegin++; //--Выход, если ошибка обмена, или переход на следующий этап
  //----ИНИЦИАЛИЗАЦИЯ ПОДТЯЖКИ РЕГИСТРОВ ВНУТРЕННИМИ РЕЗИСТОРАМИ 100к(для ввода)
  case 4: z.data[0x00]=MCP_GPPUA; 							//--Адрес регистра, с которого будут записаны четыре параметра.
		  if(z.gppuA) 	z.data[1]=0xff;	else z.data[1]=0;	//--Установка направление регистра A (ввод/вывод). 
		  if(z.gppuB) 	z.data[2]=0xff;	else z.data[2]=0;	//--Установка направление регистра В (ввод/вывод).
		  Wire.beginTransmission(z.addr); Wire.write(z.data,3);	//--Передача в i2c-slave 3 байта.
		  if(Wire.endTransmission()) z.device=0; else z.device=3; z.stepBegin=0; return 1;//--Выход с ошибкой или нормальный и завершение.
} return 0;}; //--End	
//-------------------------------------------------------------------------------------------------
//            			 ОБМЕН С УСТРОЙСТВОМ
//uint8_t  dirA:1, dirB:1, ipolA:1,ipolB:1, gppuA:1, gppuB:1;    
//-------------------------------------------------------------------------------------------------
bool RT_HW_MCP23017:: device(RT_HW_MCP23017_DEVICE &z, uint8_t &outA, uint8_t &outB){
if(z.device!=3) {z.run=3; z.stepDevice=0;	return 1;}	if (!z.run) return 1;	//--Проверка готовности дисплея(device=3) и команды вывода(run>0).
//-------------------------------------------------------------------------------------------------
switch(z.stepDevice){default: z.stepDevice=0;						
	//----ПРОВЕРКА ВНЕШНЕГО ФЛАГА ЗАПУСКА РАБОТЫ С УСТРОЙСТВОМ ДЛЯ НАЧАЛА ОБМЕНА С УСТРОЙСТВОМ 
	case 0:	z.rdA=z.dirA; z.rdB=z.dirB; z.wrA=!z.dirA; z.wrB=!z.dirB;
			if((z.run==1) && (z.agoA==outA)) z.wrA=0;  
			if((z.run==1) && (z.agoB==outB)) z.wrB=0;  	
//Serial.println("run="+ String(z.run)+" rdA,rdB,wrA,wrB="+String(z.rdA)+String(z.rdB)+String(z.wrA)+String(z.wrB)+" A="+String(outA,HEX)+" B="+String(outB,HEX)+" agoA="+String(z.agoA,HEX)+" agoB="+String(z.agoB,HEX));
			z.stepWork=0; z.stepDevice++;
	case 1: if(!write(z, outA, outB)) return 0; else {z.stepWork=0;  z.stepDevice++;}//--Запись
	case 2: if(!read (z))             return 0; else {z.stepWork=0;  z.stepDevice++;}	//--Чтение	
	case 3: z.run=0; z.stepDevice=0;  return 1;} //--Завершение обмена
return 0;};
//-------------------------------------------------------------------------------------------------
//                           ЗАПИСЬ В РЕГИСТРЫ A,B (одновременный вызов с read() не допускается
//-------------------------------------------------------------------------------------------------
bool RT_HW_MCP23017:: write(RT_HW_MCP23017_DEVICE &z, uint8_t &outA, uint8_t &outB){	
	//--НАСТРОЙКА ТРАНЗАКЦИИ (первый адрес записываемого регистра, значение регистров(одно или два), кол-во записываемых байт) 
	if( z.wrA &&  z.wrB) {z.data[0]=MCP_GPIOA; if(z.ipolA) {z.data[1]=0xFF-outA;} else {z.data[1]=outA;} 
											   if(z.ipolB) {z.data[2]=0xFF-outB;} else {z.data[2]=outB;} z.qntData=2;} //--Транзакция для записи регистров A и B
	if( z.wrA && !z.wrB) {z.data[0]=MCP_GPIOA; if(z.ipolA) {z.data[1]=0xFF-outA;} else {z.data[1]=outA;} z.qntData=1;}	//--Транзакция для записи в регистр A
	if(!z.wrA &&  z.wrB) {z.data[0]=MCP_GPIOB; if(z.ipolB) {z.data[1]=0xFF-outB;} else {z.data[1]=outB;} z.qntData=1;}	//--Транзакция для записи в регистр B
	if(!z.wrA && !z.wrB) {return 1;}  											//--Выход, если ни один из флагов записи не взведен.
	Wire.beginTransmission(z.addr); Wire.write(z.data,(z.qntData+1));			//--Отправка транзакции в устройство
	if(Wire.endTransmission()) {z.device=0; return 1;}
	z.agoA=outA; z.agoB=outB;
	return 1;}						//--Завершение записи. При ошибке 0->z.device		
//-------------------------------------------------------------------------------------------------
//                           ЧТЕНИЕ В РЕГИСТРЫ A,B
//-------------------------------------------------------------------------------------------------
bool RT_HW_MCP23017:: read(RT_HW_MCP23017_DEVICE &z){
	//--НАСТРОЙКА ТРАНЗАКЦИИ (первый адрес записываемого регистра) 
	if( z.rdA &&  z.rdB) {z.data[0]=MCP_GPIOA; z.qntData=2;}  	//--Транзакция для чтения регистров A и B							
	if( z.rdA && !z.rdB) {z.data[0]=MCP_GPIOA; z.qntData=1;}	//--Транзакция для вывода регистра A
	if(!z.rdA &&  z.rdB) {z.data[0]=MCP_GPIOB; z.qntData=1;}	//--Транзакция для вывода регистра B
	if(!z.rdA && !z.rdB) {return 1;}							//--Выход, если ни один из флагов ввода не взведен.		
	//----ЗАПИСЬ АДРЕСА ЧТЕНИЯ ДАННЫХ
	Wire.beginTransmission(z.addr); Wire.write(z.data[0]); //--Запись адреса в устройство
	if(Wire.endTransmission())	{z.device=0; z.stepWork=0; return 1;}	//--Завершение отправки адреса. При ошибке выход и 0->z.device	
	//----ОТПРАВКА ЗАПРОСА НА ЧТЕНИЕ
	Wire.requestFrom(z.addr,z.qntData); 	//--Запрос на чтение регистров
	z.timeStart=millis(); z.stepWork++;	//--Установка таймера и переход на следующий этап
	//-----ОЖИДАНИЕ ПРИХОДА ДВУХ БАЙТОВ С ТАЙМ_АУТОМ
	do{if(getPastTime(z.timeStart,millis()) > DEVICE_MAX_TIME_READ) {z.device=0; return 1;}} //--Если тайм-аут превышен, завершение по ошибке  
	   while(Wire.available() < z.qntData);  
	//-----ЧТЕНИЕ РЕГИСТРОВ
	if(z.rdA) z.inA=Wire.read(); else z.inA=z.agoA;
	if(z.rdB) z.inB=Wire.read(); else z.inB=z.agoB;
return 1;}				//--Нормальный выход	
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++	
//-------------------------------------------------------------------------------------------------
//                  ПОИСК УСТРОЙСТВА НА ШИНЕ i2c 
// device: =0 нет устройства; =1 после поиска устройство не обнаружено; =2 устройство обнаружено и доступно к работе.
//-------------------------------------------------------------------------------------------------
void RT_HW_MCP23017:: find (uint8_t &step, uint8_t &addr, uint8_t &device, uint8_t &cnt, uint32_t &timeStart) {//uint8_t code;
switch(step){default: step=0;  
  case 0: switch(device){default: step=0; case 0: step=5; break; case 1: step=1; break;  case 2: break;}	 break; //--Определение точки входа.	 
  //--Поиск, если ранее устройство не было обнаружено--------------------------------------------------------------------------------       
  case 1: timeStart=millis(); step=2;
  case 2: if(getPastTime(timeStart, millis()) < DEVICE_I2C_DELAY_RESTART_LONG)  break;  //--Длинная пауза 
          Wire.beginTransmission(addr); if(Wire.endTransmission()) {device=1;   step=0;} else {timeStart=millis(); step=3;} break;  //--Переход на нормальный поиск или на начало
  case 3: if(getPastTime(timeStart, millis()) < DEVICE_I2C_DELAY_RESTART_SHORT)	break;  //--Короткая пауза перед контрольным поиском 
  case 4: Wire.beginTransmission(addr); if(Wire.endTransmission()) {device=1;   step=0;} else {device=2; step=0; }          break;  //--Завершение поиска     
  //--Первичный поиск (ранее поиск устройства не осуществлялся)----------------------------------------------------------------------                   
  case 5: Wire.beginTransmission(addr); if(Wire.endTransmission()) {            step=7;} else {timeStart=millis(); step=6;} break;  //--Первичный поиск
  case 6: if(getPastTime(timeStart, millis()) < DEVICE_I2C_DELAY_RESTART_SHORT)	break;  //--Короткая пауза перед контрольным поиском
          Wire.beginTransmission(addr); if(Wire.endTransmission()) {            step=7;} else {device=2; step=0;}           break;  //--Продолжение или успешное завершение поиска
  //--Обработка ошибок при первичном поиске-----------------------------------------------------------------------------------------
  case 7: cnt++; if(cnt >=DEVICE_I2C_QNT_RESTART){device=1; step=0; break;} timeStart=millis();step=8;	//--Завершение или подготовка  к повторному поиску
  case 8: if(getPastTime(timeStart,millis()) < DEVICE_I2C_DELAY_RESTART_MIDDLE) break; step=5; break;}}	//--Пауза для повторного цикла   
//---------------------------------------------------------------------------------------
//         ВЫЧИСЛЕНИЕ РАЗНИЦЫ МЕЖДУ ДВУМЯ ВРЕМЕННЫМИ ТОЧКАМИ
//---------------------------------------------------------------------------------------
uint32_t RT_HW_MCP23017:: getPastTime (uint32_t &t1, uint32_t t2){if (t2>=t1) {return (t2-t1);} else {return (0xFFFFFFFF-t1+t2);}};  
//---------------------------------------------------------------------------------------
//            ПЕРИОДИЧЕСКОЕ (ms) ВЫПОЛНЕНИЕ ДЕЙСТВИЙ: рекомендуется инсталлировать t1=millis();
//---------------------------------------------------------------------------------------
 bool RT_HW_MCP23017:: event_ms  (uint32_t &t1,uint32_t val){if(getPastTime (t1,millis())>=val) {t1+=val; return 1;} return 0;};
 bool RT_HW_MCP23017:: event_mcs (uint32_t &t1,uint32_t val){if(getPastTime (t1,micros())>=val) {t1+=val; return 1;} return 0;}; 
//---------------------------------------------------------------------------------------
//       УСТАНОВКА ПАРАМЕТРА id.run
//---------------------------------------------------------------------------------------
void RT_HW_MCP23017::makeRun_device (RT_HW_MCP23017_DEVICE &z,bool en) {
if(en){if(!z.en){z.en=1;z.run=3;z.count=0;} else {if(z.event){z.count++; if(z.count>=z.period){z.count=0; if(!z.run) z.run=1;}}}} else {z.en=0;}} ;
  
//---------------------------------------------------------------------------------------		
RT_HW_MCP23017 RT_HW_mcp23017;//--Создание внешнего объекта	
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++