/**************************************************************************************************
 *       БИБЛИОТЕКА ДЛЯ РАСШИРИТЕЛЯ ВВОДА/ВЫВОДА
 *					mcp23017   
 ***********************************************************
  Написана ecoins (ecoins@mail.ru) 
   2019-01-21: Начальная версия
   2018-08-12: Оптимизированая под работу  с FLProg
   2019-06-10: Полностью изменен подход
   2019_07_18: Добавлено чтение входов
   2019_07_30: Полностью переписана
   2019_09_30: Выполнена как автономная библиотека (без использования библиотеки RT_HW_MAIN.h)
 ******************************************************************/
#ifndef RT_HW_MCP23017_h
#define RT_HW_MCP23017_h

#if defined(ARDUINO) && (ARDUINO >= 100)
#include <Arduino.h>
#else
#include <WProgram.h>
#endif
//-------------------------------------------------------------------------------------------------
#include <Wire.h>

#define RT_HW_MCP23017_VER 21			//--НОМЕР ВЕРСИИ
//#define RT_HW_MCP23017_DEVICE_MAX 9		//--Максимальное кол-во устройств
//****************************************************************************************
//		КЛАСС: RT_HW_MCP23017_ID_DEVICE  - параметры устройства mcp23017
// dir:	   =0 устройство не иницил.; =1 работа устройством заблокирована (например недопустимый адрес); =2 - режим обмена с устройством. 
// device: =0 устройство не готово;	 =1 не обнаружено (периодически ищется);	=2 обнаружено и инициал. функцией begin();	=3 готово к использованию.
// run:    =0 нет запроса;			 =1 вывод по изменении;	=2 безусловный вывод;			
//*******************************	*********************************************************
class  RT_HW_MCP23017_DEVICE {  		  
#pragma pack(push, 4)			//--Установка выравнивания до 4 байт
public: 
uint8_t  addr,dir,device,run,event,data[5],qntData; 
uint8_t  stepFind,stepBegin,stepDevice,stepWork,count,workEvent; 
uint16_t  dirA:1,gppuA:1,ipolA:1,wrA:1,rdA:1,  dirB:1,gppuB:1,ipolB:1,wrB:1,rdB:1, readYes:1,writeYes:1,en:1;  
uint8_t  agoA,agoB,inA,inB;
int16_t period;
uint32_t timeStart,timeStartEvent;
#pragma pack(pop)				//--Восстановление предыдущей настройки
RT_HW_MCP23017_DEVICE(){dir=device=event=en=0;};};   //--Конструктор
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//		КЛАСС: RT_HW_MCP23017_ID_DATA	- данные устройства mcp23017
// dir:	   =0 устройство не иницил.; =1 работа устройством заблокирована (например недопустимый адрес); =2 - режим обмена с устройством. 
// device: =0 устройство не готово;	 =1 не обнаружено (периодически ищется);	=2 обнаружено и инициал. функцией begin();	=3 готово к использованию.
// run:    =0 нет запроса;			 =1 вывод по изменении;	=2 безусловный вывод;		
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//class RT_HW_MCP23017_DATA{public:  uint8_t num,dir,run,stepDevice; char mode; uint16_t period,count; uint8_t ago; uint8_t data[4];  RT_HW_MCP23017_ID() {dir=run=0;}};	
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//		КЛАСС: RT_HW_MCP23017 	- функции управления устройством mcp23017
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class RT_HW_MCP23017{
public:
// ****** Методы ******
RT_HW_MCP23017(){}; 							//--Конструктор
void 	setID_device(RT_HW_MCP23017_DEVICE &z, int16_t addr=-1, char modeA='U', char modeB='U', int16_t period=-1); //--Установка параметров устройства                     
bool    begin 		(RT_HW_MCP23017_DEVICE &z);													   //--Инициализация устройства
bool    device		(RT_HW_MCP23017_DEVICE &z, uint8_t &outA, uint8_t &outB); 												   //--Обмен данными с устройством
void	find (uint8_t &step, uint8_t &addr, uint8_t &device, uint8_t &cnt, uint32_t &timeStart);   //--Поиск устройства
//--------------------------------------------
bool    write		(RT_HW_MCP23017_DEVICE &z, uint8_t &outA, uint8_t &outB); 	//--Вывод на пины значения регистров id.GPIOA, idGPIOB;
bool    read		(RT_HW_MCP23017_DEVICE &z);		//--Чтение из устройтсва
void 	makeRun_device (RT_HW_MCP23017_DEVICE &z,bool en);
//void 	setPeriod	(RT_HW_MCP23017_ID &id, int32_t extPeriod); 
//bool	getNoFind	(RT_HW_MCP23017_ID &id){return id.noFind;};	//--Возвращает флаг - устройство не обнаружено
//bool	getLink		(RT_HW_MCP23017_ID &id){return id.link;};	//--Возвращает флаг - связь с устройством  установлена	
//bool	getDevice	(RT_HW_MCP23017_ID &id){return id.link;};	//--Возвращает флаг - устройство готово к работе 	
//bool	getInitID	(RT_HW_MCP23017_ID &id){return id.init;};	//--Возвращает флаг - параметры устройства инициализированы
//bool 	checkOutA	(RT_HW_MCP23017_ID &id){if(id.agoOutA==id.outA) return 0; id.agoOutA=id.outA; return 1;};
//bool 	checkOutB	(RT_HW_MCP23017_ID &id){if(id.agoOutB==id.outB) return 0; id.agoOutB=id.outB; return 1;};
//bool 	checkInA	(RT_HW_MCP23017_ID &id){if(id.agoInA ==id.inA)  return 0; id.agoInA =id.inA;  return 1;};
//bool 	checkInB	(RT_HW_MCP23017_ID &id){if(id.agoInB ==id.inB)  return 0; id.agoInB =id.inB;  return 1;};
//bool 	checkDevice	(RT_HW_MCP23017_ID &id){if(id.agoDevice ==id.device)  return 0; id.agoDevice =id.device;  return 1;};
//char 	getMode(char mode);

//--------------------------------------------------------------------------------------------------
uint32_t getPastTime(uint32_t &t1, uint32_t t2);		//--ВЫЧИСЛЕНИЕ РАЗНИЦЫ МЕЖДУ ДВУМЯ ВРЕМЕННЫМИ ТОЧКАМИ 
bool 	 event_ms	(uint32_t &t1, uint32_t val);	    //--ПЕРИОДИЧЕСКОЕ (ms) ВЫПОЛНЕНИЕ ДЕЙСТВИЙ: рекомендуется инсталлировать t1=millis();
bool 	 event_mcs	(uint32_t &t1, uint32_t val);	    //--ПЕРИОДИЧЕСКОЕ (ms) ВЫПОЛНЕНИЕ ДЕЙСТВИЙ: рекомендуется инсталлировать t1=millis();
};  //---END 
//-------------------------------------------------------------
//==============  Конец класса ======================================================
extern RT_HW_MCP23017 RT_HW_mcp23017; //--Создание объекта LCD типа HD44780
//===================================================================================
#endif //---RT_HW_MCP23017_h


